/*
 *  Copyright (c) 2009 Luca Abeni
 *
 *  This is free software; see gpl-3.0.txt
 */

#include <libavutil/avstring.h>
#include <libavformat/avformat.h>

#include "output.h"

static void codec_init_from(AVCodecContext *c, const AVCodecContext *orig, AVRational fps)
{
    c->codec_id = orig->codec_id;
    c->bit_rate = orig->bit_rate;
    c->extradata= orig->extradata;
    c->extradata_size= orig->extradata_size;
    c->time_base = orig->time_base;
    c->time_base.den = /*orig->time_base*/ fps.num;
    c->time_base.num = /*orig->time_base*/ fps.den;
    c->sample_rate = orig->sample_rate;
    c->channels = orig->channels;
    c->frame_size = orig->frame_size;
    if((orig->block_align == 1) && (orig->codec_id == CODEC_ID_MP3)) {
        c->block_align= 0;
    } else {
        c->block_align = orig->block_align;
    }

    c->pix_fmt = orig->pix_fmt;
    c->width = orig->width;
    c->height = orig->height;
    c->has_b_frames = orig->has_b_frames;
    c->max_b_frames = orig->max_b_frames;
}


void close_output(void *h)
{
    AVFormatContext *s = h;

    av_write_trailer(s);
    if (s->pb) {
        url_fclose(s->pb);
    }

    av_free(s->streams[0]->codec);
    av_free(s->streams[0]);

fprintf(stderr, "Stream closed!\n");

    av_free(s);
}


void *alloc_output(const char *fname)
{
    AVFormatContext *s;
    AVOutputFormat *out_format;

    s = av_alloc_format_context();
    out_format = guess_format(NULL, fname, NULL);
    if (!out_format) {
        fprintf(stderr, "Unable for find the output format\n");

        return NULL;
    }

    s->oformat = out_format;
    av_strlcpy(s->filename, fname, sizeof(s->filename));
    av_set_parameters(s, NULL);

    return s;
}

int prepare_output(void *h, const void *oh, int s_id)
{
    AVFormatContext *s = h;
    const AVFormatContext *os = oh;
    AVStream *st;
    AVCodecContext *c;
    AVCodecContext *orig = os->streams[s_id]->codec;

    st = av_new_stream(s, s->nb_streams);
    if (!st) {
        fprintf(stderr, "Cannot allocate stream\n");

        return -1;
    }

    c = st->codec;
    avcodec_get_context_defaults2(c, orig->codec_type);
//    c->codec_type = orig->codec_type;

    codec_init_from(c, orig, os->streams[s_id]->r_frame_rate);
    st->time_base = os->streams[s_id]->time_base; 
    st->r_frame_rate = os->streams[s_id]->r_frame_rate; 
    if (st->codec->codec_type == CODEC_TYPE_AUDIO) {
        st->stream_copy = 1;
        os->streams[s_id]->stream_copy = 1;
        //c->bit_rate = st->codec->bit_rate;
        c->block_align = /*orig->block_align*/ 0;
        return s->nb_streams;
    }

    /* FIXME: fill all the fields */
    c->codec_id = CODEC_ID_RAWVIDEO;
    if (s->oformat->flags & AVFMT_GLOBALHEADER) {
        c->flags|= CODEC_FLAG_GLOBAL_HEADER;
    }

    return s->nb_streams;
}

int finalize_output(void *h)
{
    AVFormatContext *s = h;
    int res;
    int i;

    for (i = 0; i < s->nb_streams; i++) {
        if (!s->streams[i]->stream_copy) {
            AVCodec *c;

            c = avcodec_find_encoder(s->streams[i]->codec->codec_id);
            if (c == NULL) {
                /* FIXME: Properly clear things??? */
                fprintf(stderr, "Error finding codec for out stream %d\n", i);

                return -1;
            }
            res = avcodec_open(s->streams[i]->codec, c);
            if (res < 0) {
                /* FIXME: Properly clear things??? */
                fprintf(stderr, "Error opening codec for out stream %d\n", i);
                fprintf(stderr, "%dx%d %x\n", s->streams[i]->codec->width, s->streams[i]->codec->height, s->streams[i]->codec->pix_fmt);

                return -1;
            }
        }
        fprintf(stderr, "TB[%d]: %d/%d\n", i, s->streams[i]->time_base.num, s->streams[i]->time_base.den);
    }

    if (!(s->oformat->flags & AVFMT_NOFILE)) {
        /* open the file */
        if (url_fopen(&s->pb, s->filename, URL_WRONLY) < 0) {
            fprintf(stderr, "Cannot open '%s'\n", s->filename);

            return -1;
        }
    }

    av_set_parameters(s, NULL);
    res = av_write_header(s);
    if (res < 0) {
      fprintf(stderr, "Cannot Initialize output\n");

      return -1;
    }
    dump_format(s, 0, s->filename, 1);
//    fprintf(stderr, "TB[%d]: %d/%d\n", 1, s->streams[1]->time_base.num, s->streams[1]->time_base.den);

    return 0;
}

int write_output(void *s, void *pkt)
{
  AVPacket *p = pkt;
  p->pts = AV_NOPTS_VALUE;
  p->dts = AV_NOPTS_VALUE;
  return av_interleaved_write_frame(s, pkt);
}
